////////////////////////////////////////////////////////////////////////////
//                           **** WAVPACK ****                            //
//                  Hybrid Lossless Wavefile Compressor                   //
//              Copyright (c) 1998 - 2006 Conifer Software.               //
//                          All Rights Reserved.                          //
//      Distributed under the BSD Software License (see license.txt)      //
////////////////////////////////////////////////////////////////////////////

// wavpack_version.h

#ifndef WAVPACK_VERSION_H
#define WAVPACK_VERSION_H

#define LIBWAVPACK_MAJOR 4
#define LIBWAVPACK_MINOR 60
#define LIBWAVPACK_MICRO 1
#define LIBWAVPACK_VERSION_STRING "4.60.1"

#endif
