#ifndef __WINZIP_JPEG_INPUT_STREAM_H__
#define __WINZIP_JPEG_INPUT_STREAM_H__

#include <stdint.h>
#include <stdlib.h>

typedef size_t WinZipJPEGReadFunction(void *context,uint8_t *buffer,size_t length);

#endif

